@@ -1,128 +1,29 @@
-"""Circuit graph/expression library.
-
-Minimal native Python library for building and working with
-logical circuits.
 """
-
+Minimal native Python library for building and working with
+logical circuits (both as expressions and as graphs).
+"""
 from __future__ import annotations
-from typing import Sequence
+from typing import Sequence, Optional, Union, Callable
 import doctest
-from math import log2
-from itertools import product
 from parts import parts
+from logical import logical
 
-class operation(tuple):
-    """
-    The list of binary logical operations represented as
-    the output columns of truth tables where the input
-    column pairs are sorted in ascending order:
-    * (0,0,0,0) is FALSE
-    * (0,0,0,1) is AND
-    * (0,0,1,0) is NIMP (i.e., >)
-    * (0,0,1,1) is FST (first/left-hand input)
-    * (0,1,0,0) is NIF (i.e., <)
-    * (0,1,0,1) is SND (second/right-hand input)
-    * (0,1,1,0) is XOR (i.e., !=)
-    * (0,1,1,1) is OR
-    * (1,0,0,0) is NOR
-    * (1,0,0,1) is XNOR (i.e., ==)
-    * (1,0,1,0) is NSND (negation of second input)
-    * (1,0,1,1) is IF (i.e., >=)
-    * (1,1,0,0) is NFST (negation of first input)
-    * (1,1,0,1) is IMP (i.e., <=)
-    * (1,1,1,0) is NAND
-    * (1,1,1,1) is TRUE
-    """
+# Synonyms (also exported).
+operation = logical
+op = logical
 
-    names = {
-        (0,1): 'id',
-        (1,0): 'not',
-        (0,0,0,0): 'false',
-        (0,0,0,1): 'and',
-        (0,0,1,0): 'nimp',
-        (0,0,1,1): 'fst',
-        (0,1,0,0): 'nif',
-        (0,1,0,1): 'snd',
-        (0,1,1,0): 'xor',
-        (0,1,1,1): 'or',
-        (1,0,0,0): 'nor',
-        (1,0,0,1): 'xnor',
-        (1,0,1,0): 'nsnd',
-        (1,0,1,1): 'if',
-        (1,1,0,0): 'nfst',
-        (1,1,0,1): 'imp',
-        (1,1,1,0): 'nand',
-        (1,1,1,1): 'true'
-    }
-
-    def __call__(self: operation, *arguments) -> int:
-        """
-        Apply the operator to an input tuple.
-
-        >>> operation((1,0))(1)
-        0
-        >>> operation((1,0,0,1))(0,0)
-        1
-        >>> operation((1,0,0,1))(1,1)
-        1
-        >>> operation((1,0,0,1))(1,0)
-        0
-        >>> operation((1,0,0,1))(0,1)
-        0
-        >>> operation((1,0,0,1,0,1,0,1))(1,1,0)
-        0
-        """
-        if len(arguments) == 1:
-            return self[[0, 1].index(arguments[0])]
-        elif len(arguments) == 2:
-            return self[[(0,0),(0,1),(1,0),(1,1)].index(tuple(arguments))]
-        else:
-            inputs = list(product(*[(0,1)]*self.arity()))
-            return self[inputs.index(tuple(arguments))]
-
-    def name(self: operation) -> str:
-        """
-        Typical name for the operator.
-
-        >>> operation((1,0,0,1)).name()
-        'xnor'
-        """
-        return dict(operation.names)[self]
-
-    def arity(self) -> int:
-        """Arity of the operator."""
-        return int(log2(len(self)))
-
-# Concise synonyms for common operations.
-operation.id_ = operation((0,1))
-operation.not_ = operation((1,0))
-operation.false_ = operation((0,0,0,0))
-operation.and_ = operation((0,0,0,1))
-operation.nimp_ = operation((0,0,1,0))
-operation.fst_ = operation((0,0,1,1))
-operation.nif_ = operation((0,1,0,0))
-operation.snd_ = operation((0,1,0,1))
-operation.xor_ = operation((0,1,1,0))
-operation.or_ = operation((0,1,1,1))
-operation.nor_ = operation((1,0,0,0))
-operation.xnor_ = operation((1,0,0,1))
-operation.nsnd_ = operation((1,0,1,0))
-operation.if_ = operation((1,0,1,1))
-operation.nfst_ = operation((1,1,0,0))
-operation.imp_ = operation((1,1,0,1))
-operation.nand_ = operation((1,1,1,0))
-operation.true_ = operation((1,1,1,1))
-
-# Concise synonym for class.
-op = operation
-
-class gate():
+class gate:
     """
     Data structure for an individual circuit logic gate.
-    """
 
+    :param operation: Logical operation that the gate represents.
+    :param inputs: List of input gate object references.
+    :param outputs: List of output gate object references.
+    :param is_input: Flag indicating if this is an input gate for a circuit.
+    :param is_output: Flag indicating if this is an output gate for a circuit.
+    """
     def __init__(
-            self: gate, operation: operation = None,
+            self: gate, operation: op = None,
             inputs: Sequence[gate] = None, outputs: Sequence[gate] = None,
             is_input: bool = False, is_output: bool = False
         ):
@@ -140,7 +41,20 @@ class gate():
             ig.output(self)
 
     def output(self: gate, other: gate):
-        """Designate another gate as an output gate of this gate."""
+        """
+        Designate another gate as an output gate of this gate.
+
+        :param other: Gate to be designated as an output gate of this gate.
+
+        >>> c = circuit()
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.and_, [g0, g1])
+        >>> g3 = c.gate(op.id_, [g2], is_output=True)
+        >>> g2.output(g3) # Confirm this is idempotent.
+        >>> c.count()
+        4
+        """
         if not any(o is other for o in self.outputs):
             self.outputs = self.outputs + [other]
 
@@ -148,29 +62,61 @@ class gates(list):
     """
     Data structure for a gate collection that appears in a circuit.
     """
-
     @staticmethod
     def mark(g: gate):
-        """Mark all gates reachable from the input gate."""
+        """
+        Mark all gates reachable from the supplied gate via recursive traversal
+        of input gate references.
+
+        :param g: Gate from which to mark all reachable gates (via input references).
+
+        >>> c = circuit()
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.and_, [g0, g1])
+        >>> g3 = c.gate(op.id_, [g2], is_output=True)
+        >>> gates.mark(g3)
+        >>> all(g.is_marked for g in [g0, g1, g2, g3])
+        True
+        """
         if not g.is_marked:
             g.is_marked = True
             for ig in g.inputs:
                 gates.mark(ig)
 
     def __call__(
-            self: gates, operation: operation = None,
+            self: gates, operation: op = None,
             inputs: Sequence[gate] = None, outputs: Sequence[gate] = None,
             is_input: bool = False, is_output: bool = False
         ):
-        """Add a gate to this collection of gates."""
+        """
+        Add a gate to this collection of gates.
+
+        :param operation: Logical operation that the gate represents.
+        :param inputs: List of input gate object references.
+        :param outputs: List of output gate object references.
+        :param is_input: Flag indicating if this is an input gate for a circuit.
+        :param is_output: Flag indicating if this is an output gate for a circuit.
+
+        >>> gs = gates([])
+        >>> g0 = gs(op.id_, is_input=True)
+        >>> g1 = gs(op.id_, is_input=True)
+        >>> g2 = gs(op.and_, [g0, g1])
+        >>> g3 = gs(op.id_, [g2], is_output=True)
+        >>> len(gs)
+        4
+        """
         g = gate(operation, inputs, outputs, is_input, is_output)
         g.index = len(self)
         self.append(g)
         return g
 
-class signature():
+class signature:
     """
-    Data structure for a circuit signatures.
+    Data structure for a circuit signature.
+
+    :param input_format: List of bit vector lengths of inputs.
+    :param output_format: List of bit vector lengths of outputs.
 
     >>> s = signature()
     >>> s.input([1, 2, 3])
@@ -186,7 +132,6 @@ class signature():
       ...
     TypeError: signature output format must be a list of integers
     """
-
     def __init__(
             self: signature,
             input_format: Sequence[int] = None,
@@ -206,10 +151,16 @@ class signature():
             raise TypeError('signature output format must be a list of integers')
         self.output_format = output_format
 
-    def input(self: signature, input):
+    def input(self: signature, input: Sequence[Sequence[int]]) -> Sequence[int]:
         """
         Convert an input organized in a way that matches the
         signature's input format into a flat list of bits.
+
+        :param input: Input bit vector that matches signature.
+
+        >>> s = signature(input_format=[2, 3])
+        >>> s.input([[1, 0], [0, 1, 1]])
+        [1, 0, 0, 1, 1]
         """
         if self.input_format is None:
             return input
@@ -224,10 +175,16 @@ class signature():
         else:
             raise ValueError('input format does not match signature')
 
-    def output(self: signature, output):
+    def output(self: signature, output: Sequence[int]) -> Sequence[Sequence[int]]:
         """
         Convert a flat list of output bits into a format that
         matches the signature's output format specification.
+
+        :param output: Flat output bit vector to convert (according to signature).
+
+        >>> s = signature(output_format=[2, 3])
+        >>> list(s.output([1, 0, 0, 1, 1]))
+        [[1, 0], [0, 1, 1]]
         """
         if self.output_format is None:
             return output
@@ -238,6 +195,8 @@ class circuit():
     """
     Data structure for an instance of a circuit.
 
+    :param sig: Signature (input and output bit vector lengths) for the circuit.
+
     >>> c = circuit()
     >>> c.count()
     0
@@ -245,17 +204,29 @@ class circuit():
     >>> g1 = c.gate(op.id_, is_input=True)
     >>> g2 = c.gate(op.and_, [g0, g1])
     >>> g3 = c.gate(op.id_, [g2], is_output=True)
-    >>> g2.output(g3) # Confirm this is idempotent.
     >>> c.count()
     4
+
+    An instance can be evaluated on any list of bits using the :obj:`evaluate`
+    method. The result is a bit vector that includes one bit for each output
+    gate.
+
     >>> [list(c.evaluate(bs)) for bs in [[0, 0], [0, 1], [1, 0], [1, 1]]]
     [[0], [0], [0], [1]]
+
+    It is also possible to remove all internal gates from a circuit from which
+    an output gate cannot be reached. Doing so does not change the order of the
+    input gates or the order of the output gates.
+
     >>> c.prune_and_topological_sort_stable()
     >>> c.count()
     3
     >>> [list(c.evaluate(bs)) for bs in [[0, 0], [0, 1], [1, 0], [1, 1]]]
     [[0], [0], [0], [1]]
 
+    It is also possible to specify the signature of a circuit using the
+    :obj:`signature` class.
+
     >>> c = circuit(signature([2], [1]))
     >>> c.count()
     0
@@ -274,22 +245,27 @@ class circuit():
     5
     >>> [list(c.evaluate([bs])) for bs in [[0, 0], [0, 1], [1, 0], [1, 1]]]
     [[[0]], [[1]], [[1]], [[0]]]
-    >>> c.evaluate([[0, 0, 0]])
-    Traceback (most recent call last):
-      ...
-    ValueError: input format does not match signature
-    >>> c.evaluate([0, 0])
-    Traceback (most recent call last):
-      ...
-    TypeError: input must be a list of integer lists
     """
-
-    def __init__(self: circuit, sig: signature = None):
+    def __init__(self: circuit, sig: Optional[signature] = None):
         self.gate = gates([])
         self.signature = signature() if sig is None else sig
 
-    def count(self: circuit, predicate=lambda _: True) -> int:
-        """Count the number of gates that satisfy the supplied predicate."""
+    def count(self: circuit, predicate: Callable[[gate], bool] = lambda _: True) -> int:
+        """
+        Count the number of gates that satisfy the supplied predicate.
+
+        :param predicate: Function that distinguishes certain gate objects.
+
+        >>> c = circuit(signature([2], [1]))
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.not_, [g0])
+        >>> g3 = c.gate(op.not_, [g1])
+        >>> g4 = c.gate(op.xor_, [g2, g3])
+        >>> g5 = c.gate(op.id_, [g4], is_output=True)
+        >>> c.count(lambda g: g.operation == op.id_)
+        3
+        """
         return len([() for g in self.gate if predicate(g)])
 
     def prune_and_topological_sort_stable(self: circuit):
@@ -298,6 +274,19 @@ class circuit():
         and topologically sort the gates (with input gates all in
         their original order at the beginning and output gates all
         in their original order at the end).
+
+        >>> c = circuit(signature([2], [1]))
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.not_, [g0])
+        >>> g3 = c.gate(op.not_, [g1])
+        >>> g4 = c.gate(op.xor_, [g2, g3])
+        >>> g5 = c.gate(op.id_, [g4], is_output=True)
+        >>> c.count()
+        6
+        >>> c.prune_and_topological_sort_stable()
+        >>> c.count()
+        5
         """
 
         # Collect all gates that feed directly into the identity gates
@@ -344,11 +333,48 @@ class circuit():
 
         self.gate = gate_
 
-    def evaluate(self: circuit, input):
+    def evaluate(
+            self: circuit,
+            input: Union[Sequence[int], Sequence[Sequence[int]]]
+        ) -> Union[Sequence[int], Sequence[Sequence[int]]]:
         """
         Evaluate the circuit on an input organized in a way that
         matches the circuit signature's input format, and return
         an output that matches the circuit signature's output format.
+
+        :param input: Input bit vector or bit vectors.
+
+        >>> c = circuit()
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.and_, [g0, g1])
+        >>> g3 = c.gate(op.id_, [g2], is_output=True)
+        >>> list(c.evaluate([0, 1]))
+        [0]
+
+        It is also possible to evaluate a circuit that has a signature
+        specified. Note that in this case, the inputs and outputs must
+        be lists of lists (to reflect that there are multiple inputs).
+
+        >>> c = circuit(signature([2], [1]))
+        >>> g0 = c.gate(op.id_, is_input=True)
+        >>> g1 = c.gate(op.id_, is_input=True)
+        >>> g2 = c.gate(op.and_, [g0, g1])
+        >>> g3 = c.gate(op.id_, [g2], is_output=True)
+        >>> list(c.evaluate([[0, 1]]))
+        [[0]]
+
+        Any attempt to evaluate a circuit on an invalid input raises
+        an exception.
+
+        >>> c.evaluate([[0, 0, 0]])
+        Traceback (most recent call last):
+          ...
+        ValueError: input format does not match signature
+        >>> c.evaluate([0, 0])
+        Traceback (most recent call last):
+          ...
+        TypeError: input must be a list of integer lists
         """
         input = self.signature.input(input)
         wire = input + [None]*(self.count(lambda g: len(g.inputs) > 0))
